#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: distance.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Mon Jun 30 13:23:40 2014 CST
"""
DESCRIPTION = """
Computing distances

Common parameters:
    is_sparse:
        is the feature sparse
    is_trans:
        is `feat` transfered
"""

import logging

import numpy as np
import scipy as sp


def Euclidean(feat, query=None,
              is_sparse=False, is_trans=False):
    """ Euclidean distance.
    """
    if None == query:
        (N, D) = feat.shape
        dotprod = feat.dot(feat.T)
        featl2norm = sp.repeat(dotprod.diagonal().reshape(1, -1), N, 0)
        qryl2norm = featl2norm.T
    else:
        (nQ, D) = query.shape
        (N, D) = feat.shape
        dotprod = query.dot(feat.T)
        qryl2norm = \
            sp.repeat(np.multiply(query, query).sum(1).reshape(-1, 1), N,  1)
        featl2norm = \
            sp.repeat(np.multiply(feat, feat).sum(1).reshape(1, -1), nQ, 0)

    return qryl2norm + featl2norm - 2 * dotprod


def Euclidean_DML(feat, M, query=None,
                  is_sparse=False, is_trans=False):
    """ Euclidean distance with DML.
    """
    (N, D) = feat.shape
    dotprod = feat.dot(M).dot(feat.T)
    l2norm = sp.repeat(dotprod.diagonal().reshape(1, -1), N, 0)
    return l2norm + l2norm.T - 2 * dotprod


def Cosine(feat, query=None,
           is_sparse=False, is_trans=False):
    """ Cosine distance.
    """
    logging.info("Cosine")
    try:
        feat = feat / np.sqrt(np.sum(np.multiply(feat, feat), axis=1))\
            .reshape((feat.shape[0], 1))
    except ValueError as e:
        logging.debug("%s (take feature as sparse matrix)" % e.message)
        feat2 = feat.copy()
        feat2.data **= 2
        rows_sums = np.array(feat2.sum(axis=1))[:, 0]
        feat.data /= rows_sums[feat2.nonzero()[0]]

    if None == query:
        query = feat
    else:
        try:
            query = query/np.sqrt(np.sum(np.multiply(query, query), axis=1))\
                .reshape((query.shape[0], 1))
        except ValueError as e:
            logging.debug("%s (take query as sparse matrix)" % e.message)
            query2 = query.copy()
            query2.data **= 2
            rows_sums = np.array(query2.sum(axis=1))[:, 0]
            query.data /= rows_sums[query2.nonzero()[0]]

    return DotProduct(feat, query)


def Cosine_DML(feat, M, query=None,
               is_sparse=False, is_trans=False):
    """ Cosine distance with DML.
    """
    try:
        feat = feat/np.sqrt(np.sum(np.multiply(feat, feat), axis=1))\
            .reshape((feat.shape[0], 1))
    except ValueError as e:
        logging.debug("%s (take feature as sparse matrix)" % e.message)
        feat2 = feat.copy()
        feat2.data **= 2
        rows_sums = np.array(feat2.sum(axis=1))[:, 0]
        feat.data /= rows_sums[feat2.nonzero()[0]]

    if None == query:
        query = feat
    else:
        try:
            query = query/np.sqrt(np.sum(np.multiply(query, query), axis=1))\
                .reshape((query.shape[0], 1))
        except ValueError as e:
            logging.debug("%s(take query as sparse matrix)" % e.message)
            query2 = query.copy()
            query2.data **= 2
            rows_sums = np.array(query2.sum(axis=1))[:, 0]
            query.data /= rows_sums[query2.nonzero()[0]]

    return DotProduct_DML(feat, M, query)


def DotProduct(feat, query=None,
               is_sparse=False, is_trans=False):
    """ DotProduct distance.
    """
    logging.debug("DotProduct")
    if None == query:
        query = feat
    return -query.dot(feat.T)


def DotProduct_DML(feat, M, query=None,
                   is_sparse=False, is_trans=False):
    """ DotProduct distance with DML.
    """
    if None == query:
        query = feat
    return -query.dot(M).dot(feat.T)


def DotProduct_DML_Diagonal(feat, M, query=None,
                            is_sparse=False, is_trans=False):
    """ DotProduct distance with DML.
    """
    if None == query:
        query = feat
    query.data = query.data * M[query.nonzero()[1]]
    return -query.dot(feat.T)


def Intersection(feat, query=None,
                 is_sparse=False, is_trans=False):
    """ Intersection distance.
    """
    raise Exception("Untested function")
    if None == query:
        query = feat
    qnum = query.shape[0]
    fnum = feat.shape[0]
    dist = np.zeros((qnum, fnum))

    for i in range(qnum):
        for j in range(fnum):
            dist[i, j] = sp.vstack((query[i, :], feat[j, :])).min(0).sum()
    return dist


def Intersection_DML(feat, M, query=None,
                     is_sparse=False, is_trans=False):
    """ Intersection distance with DML.
    """
    raise Exception("Untested function")
    if None == query:
        query = feat

    qnum = query.shape[0]
    fnum = feat.shape[0]
    dist = np.zeros((qnum, fnum))

    for i in range(qnum):
        query = query[i, :].dot(M)
        for j in range(fnum):
            dist[i, j] = sp.vstack((query, feat[j, :])).min(0).sum()
    return -query.dot(M).dot(feat.T)


distFunc = {
    "euclidean":    Euclidean,
    "dotproduct":   DotProduct,
    "intersection": Intersection,
    "cosine":       Cosine}


distFunc_DML = {
    "euclidean":    Euclidean_DML,
    "dotproduct":   DotProduct_DML,
    "intersection": Intersection_DML,
    "cosine":       Cosine_DML}



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: storage.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 11:25:03 2014 CST
"""
DESCRIPTION = """
"""

import itertools
from struct import pack, unpack
import lmdb
import cPickle as pickle
import ipdb


class LMDBAccessor(object):
    def __init__(self, env, dbname):
        self.env = env
        self.db = env.open_db(dbname)

    def close(self):
        return
        self.db.close()

    def setkv(self, key, val, ktype, vtype):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(pack(ktype, key), pack(vtype, key))

    def setk(self, key, val, ktype):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(pack(ktype, key), val)

    def setv(self, key, val, vtype):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(key, pack(vtype, val))

    def set(self, key, val):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(key, val)

    def getkv(self, key, ktype, vtype):
        with self.env.begin(write=True) as txt:
            val = txt.cursor(self.db).get(pack(ktype, key))
            if val is None:
                return None
            return unpack(vtype, val)[0]

    def getk(self, key, ktype):
        with self.env.begin(write=True) as txt:
            return txt.cursor(self.db).get(pack(ktype, key))

    def getv(self, key, vtype):
        with self.env.begin(write=True) as txt:
            val = txt.cursor(self.db).get(key)
            if val is None:
                return None
            return unpack(vtype, val)[0]

    def get(self, key):
        with self.env.begin(write=True) as txt:
            return txt.cursor(self.db).get(key)

    def setkivi(self, key, val):
        self.setkv(key, val, 'i', 'i')

    def setki(self, key, val):
        self.setk(key, val, 'i')

    def setvi(self, key, val):
        self.setv(key, val, 'i')

    def getkivi(self, key):
        return self.getkv(key, 'i', 'i')

    def getki(self, key):
        return self.getk(key, 'i')

    def getvi(self, key):
        return self.getv(key, 'i')


class PQStorage(object):
    def __init__(self, pardic=None):
        self.blocks = None
        self.keys = None
        self.num_emptys = -1
        self.num_items = -1

    def get_num_items(self):
        return self.num_items

    def get_num_emptys(self):
        return self.num_emptys

    def __iter__(self):
        raise Exception("Instance of `PQStorage` is not allowed!")

    def next(self):
        raise Exception("Instance of `PQStorage` is not allowed!")

    def clear(self):
        raise Exception("Instance of `PQStorage` is not allowed!")


class MemPQStorage(PQStorage):
    def __init__(self, pardic=None):
        PQStorage.__init__(self)
        self.blocks = []
        self.keys = []
        self.num_items = 0

    def add(self, codes, keys):
        self.blocks.append(codes)
        self.keys.append(keys)
        self.num_items += codes.shape[0]

    def __iter__(self):
        return itertools.izip(self.keys, self.blocks)


class LMDBPQStorage(PQStorage):
    def __init__(self, pardic=None):
        PQStorage.__init__(self)

        path = pardic['path']
        clear = pardic.get('clear', False)

        self.env = lmdb.open(path, max_dbs=3)
        self.db_keys = LMDBAccessor(self.env, 'keys')
        self.db_vals = LMDBAccessor(self.env, 'vals')
        self.db_info = LMDBAccessor(self.env, 'info')

        if clear:
            self.clear()

        self.num_items = self.db_info.getvi('num_items')
        if self.num_items is None:
            self.num_items = 0
            self.db_info.setvi('num_items', self.num_items)

    def __del__(self):
        self.db_keys.close()
        self.db_vals.close()
        self.db_info.close()
        self.env.close()

    def add(self, codes, keys):
        key = "%08d" % self.num_items

        self.db_keys.set(key, pickle.dumps(keys, protocol=2))
        self.db_vals.set(key, pickle.dumps(codes, protocol=2))

        self.num_items += codes.shape[0]
        self.db_info.setvi('num_items', self.num_items)

    def clear(self):
        with self.env.begin(write=True) as txt:
            txt.drop(self.db_keys.db, False)
            txt.drop(self.db_vals.db, False)
            txt.drop(self.db_info.db, False)

    def __iter__(self):
        self.iter_txt = self.env.begin()
        self.cursor_keys = self.iter_txt.cursor(self.db_keys.db)
        self.cursor_vals = self.iter_txt.cursor(self.db_vals.db)
        return self

    def next(self):
        status_keys = self.cursor_keys.next()
        status_vals = self.cursor_vals.next()
        if status_keys and status_vals:
            keys = pickle.loads(self.cursor_keys.value())
            vals = pickle.loads(self.cursor_vals.value())
            return keys, vals
        else:
            # self.iter_txt.close()
            raise StopIteration


PQ_DIC = {
    'mem':  MemPQStorage,
    'lmdb': LMDBPQStorage,
}



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: util.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 09:38:24 2014 CST
"""
DESCRIPTION = """
"""

"""
KMeans
"""
# from sklearn.cluster import KMeans
# from yael import ynumpy
import cv2

# distance
from distance import distFunc
import bottleneck


"""
Different options for kmeans.
1. kmeans_yael
2. kmeans_cv2
3. kmeans_sklearn
"""

"""
def kmeans_yael(vs, ks, niter):
    return ynumpy.kmeans(vs, ks, niter=100, verbose=True)


def kmeans_sklearn(vs, ks, niter):
    kmeans = KMeans(n_clusters=ks, max_iter=100)
    kmeans.fit(vs)
    return kmeans.cluster_centers_
"""


def kmeans_cv2(vs, ks, niter):
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER,
                100, 0.01)
    flags = cv2.KMEANS_RANDOM_CENTERS
    compactness, labels, centers = cv2.kmeans(
        vs, ks, criteria, 1, flags)
    return centers


kmeans = kmeans_cv2


# finding nearest neighbor

def pq_kmeans_assign(centroids, query):
    dist = distFunc['euclidean'](centroids, query)
    return dist.argmin(1)


def pq_knn(dist, topk):
    ids = bottleneck.argpartsort(dist, topk)[:topk]
    ids = ids[dist[ids].argsort()]
    return ids



#!/usr/bin/env python
# coding: utf-8
"""
   File Name: setup.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 08:55:15 2014 CST
 Description:
"""


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print "warning: pypandoc module not found, DONOT convert Markdown to RST"
    read_md = lambda f: open(f, 'r').read()


setup(name='hdidx',
      version='0.0.3',
      author='WAN Ji',
      author_email='wanji@live.com',
      package_dir={'hdidx': 'src'},
      packages=['hdidx'],
      scripts=['tests/test_indexer.py'],
      url='https://github.com/wanji/hdidx',
      # license='LICENSE.txt',
      description='ANN Search in High-Dimensional Spaces',

      long_description=read_md("DESC.md"),
      install_requires=[
          "numpy      >= 1.6.0",
          "scipy      >= 0.9.0",
          "bottleneck >= 0.8.0",
          "lmdb       >= 0.83",
      ],
      )



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: test_indexer.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 10:38:24 2014 CST
"""
DESCRIPTION = """
"""

import unittest

import time
import logging

import numpy as np

from hdidx.distance import distFunc
from hdidx import indexer


def load_random(ntrain, nbase, nquery, d=16):
    """
    Generate a set of unit norm vectors
    """
    np.random.seed(0)
    vtrain = np.random.random((ntrain, d))
    vbase = np.random.random((nbase, d))
    vquery = np.random.random((nquery, d))

    t0 = time.clock()
    ids_gnd = np.empty(nquery)
    logging.info("Computing the ground-truth...")
    batsize = 20
    for q in range(0, nquery, batsize):
        logging.info("\r\t%d/%d" % (q, nquery))
        last = min(q+batsize, nquery)
        dist = distFunc['euclidean'](vbase, vquery[q:last])
        ids_gnd[q:last] = dist.argmin(1)
    logging.info("\r\t%d/%d\tDone!" % (nquery, nquery))
    # dis_gnd = [dist[i, ids_gnd[i]] for i in range(dist.shape[0])]
    tgnd = time.clock() - t0
    logging.info("GND Time: %.3fs" % tgnd)
    return vtrain, vbase, vquery, ids_gnd


def create_random_data():
    """
    Create random data
    """
    # synthetic dataset
    vtrain, vbase, vquery, ids_gnd = load_random(
        ntrain=10**4, nbase=10**4, nquery=10**1)

    return np.require(vtrain, np.single, requirements="C"),\
        np.require(vbase, np.single, requirements="C"),    \
        np.require(vquery, np.single, requirements="C"),   \
        ids_gnd


def compute_stats(nquery, ids_gnd, ids_pqc, k):
    # ipdb.set_trace()
    nn_ranks_pqc = np.zeros(nquery)
    qry_range = np.arange(ids_pqc.shape[1])
    for i in range(nquery):
        gnd_ids = ids_gnd[i]
        nn_pos = qry_range[ids_pqc[i, :] == gnd_ids]

        if len(nn_pos) == 1:
            nn_ranks_pqc[i] = nn_pos
        else:
            nn_ranks_pqc[i] = k + 1

    nn_ranks_pqc.sort()

    for i in [1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000]:
        if i > k:
            break
        r_at_i = (nn_ranks_pqc < i).sum() * 100.0 / nquery
        logging.warning('r@%3d = %.3f' % (i, r_at_i))


class TestPQNew(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        cls.vtrain, cls.vbase, cls.vquery, cls.ids_gnd = \
            create_random_data()

    @classmethod
    def tearDownClass(cls):
        pass

    def test_pq_mem(self):
        nsubq = 8
        topk = 100

        idx = indexer.PQIndexer()
        idx.build({
            'vals': self.vtrain,
            'nsubq': nsubq,
        })
        idx.save('mem.info')
        idx.set_storage('mem')

        idx.add(self.vbase)
        ids, dis = idx.search(self.vquery, topk=topk)
        compute_stats(self.vquery.shape[0], self.ids_gnd, ids, topk)

    def test_pq_lmdb(self):
        nsubq = 8
        topk = 100

        idx = indexer.PQIndexer()
        idx.build({
            'vals': self.vtrain,
            'nsubq': nsubq,
        })
        idx.save('lmdb.info')
        idx.set_storage('lmdb', {
            'path': 'lmdb.idx',
            'clear': True,
        })
        idx.add(self.vbase)
        ids, dis = idx.search(self.vquery, topk=topk)
        compute_stats(self.vquery.shape[0], self.ids_gnd, ids, topk)

        idx1 = indexer.PQIndexer()
        idx1.load('lmdb.info')
        idx1.set_storage('lmdb', {
            'path': 'lmdb.idx',
            'clear': True,
        })
        idx1.add(self.vbase)
        ids, dis = idx1.search(self.vquery, topk=topk)
        compute_stats(self.vquery.shape[0], self.ids_gnd, ids, topk)

        idx2 = indexer.PQIndexer()
        idx2.load('lmdb.info')
        idx2.set_storage('lmdb', {
            'path': 'lmdb.idx',
            'clear': False,
        })
        ids, dis = idx2.search(self.vquery, topk=topk)
        compute_stats(self.vquery.shape[0], self.ids_gnd, ids, topk)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: indexer.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 09:07:38 2014 CST
"""
DESCRIPTION = """
"""

# import ipdb
import cPickle as pickle
import logging

import numpy as np

from util import kmeans, pq_kmeans_assign, pq_knn
from distance import distFunc
from storage import PQ_DIC


class Indexer(object):
    class IdxData(object):
        pass

    def __init__(self):
        self.INST_ERR = "Instance of `Indexer` is not allowed!"
        pass

    def __del__(self):
        pass

    def build(self, vals=None, labels=None):
        """
        Build the indexer based on given training data
        """
        raise Exception(self.INST_ERR)

    def load(self, path):
        """
        Load indexer information from file
        """
        with open(path, 'rb') as pklf:
            self.idxdat = pickle.load(pklf)

    def save(self, path):
        """
        Save the information related to the indexer itself
        """
        with open(path, 'wb') as pklf:
            pickle.dump(self.idxdat, pklf, protocol=2)

    def set_storage(self, storage_type='mem', storage_parm=None):
        """
        Set up the backend storage engine
        """
        raise Exception(self.INST_ERR)

    def add(self, vals, keys):
        """
        Add one or more items to the indexer
        """
        raise Exception(self.INST_ERR)

    def remove(self, keys):
        """
        Remove one or more items from the indexer
        """
        raise Exception(self.INST_ERR)

    def search(self, query, topk=None, thresh=None):
        """
        Search in the indexer for `k` nearest neighbors or
        neighbors in a distance of `thresh`
        """
        raise Exception(self.INST_ERR)


class PQIndexer(Indexer):
    def __init__(self):
        self.INST_ERR = "Instance of `Indexer` is not allowed!"
        pass

    def __del__(self):
        pass

    def build(self, pardic=None):
        # training data
        vals = pardic['vals']
        # the number of subquantizers
        nsubq = pardic['nsubq']
        # the number bits of each subquantizer
        nsubqbits = pardic.get('nsubqbits', 8)
        # the number of items in one block

        # blksize = pardic.get('blksize', 16384)

        # vector dimension
        dim = vals.shape[1]
        # dimension of the subvectors to quantize
        dsub = dim / nsubq
        # number of centroids per subquantizer
        ksub = 2 ** nsubqbits

        """
        Initializing indexer data
        """
        # idxdat = Indexer.IdxData()
        # idxdat.nsubq = nsubq
        # idxdat.ksub = ksub
        # idxdat.dsub = dsub
        # idxdat.centroids = [None for q in range(nsubq)]

        idxdat = dict()
        idxdat['nsubq'] = nsubq
        idxdat['ksub'] = ksub
        idxdat['dsub'] = dsub
        idxdat['centroids'] = [None for q in range(nsubq)]

        logging.info("Building codebooks in subspaces - BEGIN")
        for q in range(nsubq):
            logging.info("\tsubspace %d/%d:" % (q, nsubq))
            vs = np.require(vals[:, q*dsub:(q+1)*dsub], requirements='C')
            idxdat['centroids'][q] = kmeans(vs, ksub, niter=100)
        logging.info("Building codebooks in subspaces - DONE")

        self.idxdat = idxdat

    def set_storage(self, storage_type='mem', storage_parm=None):
        self.storage = PQ_DIC[storage_type](storage_parm)

    def add(self, vals, keys=None):
        num_vals = vals.shape[0]

        dsub = self.idxdat['dsub']
        nsubq = self.idxdat['nsubq']
        centroids = self.idxdat['centroids']

        codes = np.zeros((num_vals, nsubq), np.uint8)
        for q in range(nsubq):
            vsub = vals[:, q*dsub:(q+1)*dsub]
            codes[:, q] = pq_kmeans_assign(centroids[q], vsub)
        self.storage.add(codes, keys)

#         start_id = 0
#         while start_id < num_vals:
#             if self.num_emptys is 0:
#                 self.blocks.append(np.zeros((blksize, nsubq), np.uint8))
#                 self.num_emptys = blksize
#             cur_num = min(self.num_emptys, num_vals-start_id)
#             cur_start = blksize - self.num_emptys
#
#             for q in range(nsubq):
#                 vsub = vals[start_id:start_id+cur_num, q*dsub:(q+1)*dsub]
#                 # ipdb.set_trace()
#                 self.blocks[-1][cur_start:cur_start+cur_num, q] = \
#                     pq_kmeans_assign(centroids[q], vsub)
#
#             self.num_emptys -= cur_num
#             start_id += cur_num
        # ipdb.set_trace()

    def remove(self, keys):
        raise Exception(self.INST_ERR)

    def search(self, query, topk=None, thresh=None):
        nq = query.shape[0]

        dsub = self.idxdat['dsub']
        nsubq = self.idxdat['nsubq']
        ksub = self.idxdat['ksub']
        centroids = self.idxdat['centroids']

        distab = np.zeros((nsubq, ksub), np.single)
        dis = np.zeros((nq, topk), np.single)
        ids = np.zeros((nq, topk), np.int)

        for qid in range(nq):
            # pre-compute the table of squared distance to centroids
            for q in range(nsubq):
                vsub = query[qid:qid+1, q*dsub:(q+1)*dsub]
                distab[q:q+1, :] = distFunc['euclidean'](
                    centroids[q], vsub)

            # add the tabulated distances to construct the distance estimators
            disquerybase = self.sumidxtab(distab)
            cur_ids = pq_knn(disquerybase, topk)

            dis[qid, :] = disquerybase[cur_ids]
            ids[qid, :] = cur_ids

        return ids, dis

    def sumidxtab(self, D):
        """
        Compute distance to database items based on distances to centroids.
            D: nsubq x ksub
        """

        nsubq = D.shape[0]
        num_base_items = self.storage.get_num_items()

        dis = np.zeros(num_base_items)
        start_id = 0

        for keys, blk in self.storage:
            cur_num = blk.shape[0]
            dis[start_id:start_id+cur_num] = \
                [sum([D[j, blk[i, j]] for j in range(nsubq)])
                 for i in range(cur_num)]
            start_id += cur_num

        return dis
